import base64
from os import urandom
from .base import BaseEncryptor
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding


class InvalidSaltError(Exception):
    """Exception raised when the provided salt does not match the original."""
    pass


class Encryptor(BaseEncryptor):

    label = 'rsa'

    @classmethod
    def generate_salt(cls) -> str:
        """Generate a random salt (32 bytes) and return it as a base64-encoded string."""
        random_bytes: bytes = urandom(32)
        return base64.b64encode(random_bytes).decode("utf-8")

    @classmethod
    def generate_secret(cls) -> str:
        """Generate and return a base64-encoded private key."""
        private_key: rsa.RSAPrivateKey = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048
        )
        return cls._make_secret_key(private_key)

    @classmethod
    def encrypt(cls, plain_text: str, secret_key: str, salt: str) -> str:
        """Encrypt plain_text with salt using the public key extracted from the secret key."""
        private_key = cls._get_private_key(secret_key)
        public_key = private_key.public_key()
        salted_text: str = f"{salt}:{plain_text}"
        encrypted_bytes = public_key.encrypt(
            salted_text.encode(),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        return base64.b64encode(encrypted_bytes).decode('utf-8')

    @classmethod
    def decrypt(cls, ciphertext: str, secret_key: str, salt: str) -> str:
        """Decrypt an RSA-encrypted string using the private key and validate the salt."""
        private_key = cls._get_private_key(secret_key)
        encrypted_bytes = base64.b64decode(ciphertext.encode('utf-8'))

        decrypted_bytes = private_key.decrypt(
            encrypted_bytes,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        salted_string = decrypted_bytes.decode('utf-8')
        extracted_salt, original_text = salted_string.split(":", maxsplit=1)
        if extracted_salt != salt:
            raise InvalidSaltError("Provided salt does not match the original salt")
        return original_text

    @classmethod
    def _make_secret_key(cls, private_key: rsa.RSAPrivateKey) -> str:
        """Convert a private key to a base64-encoded string."""
        private_bytes: bytes = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        )

        secret_key: str = base64.b64encode(private_bytes).decode('utf-8')
        return secret_key

    @classmethod
    def _get_private_key(cls, secret_key: str) -> rsa.RSAPrivateKey:
        """Extract and return the private key from a base64-encoded secret key."""
        private_bytes = base64.b64decode(secret_key.encode('utf-8'))
        private_key = serialization.load_pem_private_key(
            private_bytes,
            password=None
        )
        return private_key
    
    @classmethod
    def test(cls):
        """
        >>> Encryptor.test() # doctest: +ELLIPSIS
        Original...
        """
        secret_key = Encryptor.generate_secret()
        salt = Encryptor.generate_salt() 

        original_text = "Hello, RSA encryption!"
        encrypted_text = Encryptor.encrypt(original_text, secret_key, salt)
        decrypted_text = Encryptor.decrypt(encrypted_text, secret_key, salt)

        print("Original text:", original_text)
        print("Encrypted text:", encrypted_text)
        print("Decrypted text:", decrypted_text)

        # Verify that decryption returns the original text
        assert original_text == decrypted_text, "Decrypted text does not match the original"
        print("Test passed: Decrypted text matches the original!")




from typing import Literal, Dict, Tuple
from typing_extensions import TypeAlias

from importlib import import_module
from functools import lru_cache

from .algorithms.base import BaseEncryptor

ALGORITHM: TypeAlias = Literal["AES", "RSA"]


@lru_cache(maxsize=None)
def _get_secret_regex(sep: str = '$') -> str:
    """returns secret regex for gotten sep and available algorithms"""
    a = '|'.join(ALGORITHM.__args__)
    if sep in '.$?[({-':
        sep = rf"\{sep}"
    return rf"^({a}){sep}.+{sep}.+$"


def is_like_secret(password: str, sep: str = '$') -> bool:
    """checks whether the string seems like to be a secret"""
    import re
    return bool(re.match(_get_secret_regex(sep), password))


@lru_cache(maxsize=None)
def get_encryptor_class(alg: ALGORITHM) -> BaseEncryptor:
    """Returns encryptor class based on alg (alg=filename)"""
    try:
        module = import_module(
            name=BaseEncryptor.__module__.rsplit('.', maxsplit=1)[0] + '.' + alg.lower()
        )

        return getattr(module, "Encryptor")
    except (ImportError, AttributeError):
        raise ValueError(f"Invalid algorithm: {alg}, available: {ALGORITHM.__args__}")


def encrypt_password(password: str, alg: ALGORITHM, secret_key: str, sep: str = "$") -> str:
    """
    Make an encrypted string in defined format based on alg, secret key + generated secret key.
    Competitive with **load_passwords[varname] result
    Note: made for web only/test usage
    """
    encryptor = get_encryptor_class(alg)
    return encryptor.encrypt_full(plain_text=password, secret_key=secret_key, sep=sep)


def load_ciphertext(
    alg: ALGORITHM,
    ciphertext: str,
    secret_key: str,
    salt: str,
) -> str:
    """Decrypt ciphertext using provided attributes """
    encryptor = get_encryptor_class(alg=alg)
    decrypted_text = encryptor.decrypt(
        ciphertext=ciphertext, secret_key=secret_key, salt=salt
    )
    return decrypted_text


def decrypt_password(full_password: str, secret_key: str, sep: str = "$") -> Tuple[str, ALGORITHM]:
    """
    decrypts password string
    Args:
        full_password: password in format AES$...
        secret_key:
        sep:

    Returns:
        - decrypted string
        - using algorithm
    """

    algorithm: ALGORITHM
    try:
        assert is_like_secret(full_password, sep)
        algorithm, salt, ciphertext = full_password.split(sep, maxsplit=2)
        assert ciphertext and salt and algorithm in ALGORITHM.__args__
    except (ValueError, AssertionError):
        raise ValueError("Incorrect encrypted string format")

    try:
        return load_ciphertext(algorithm, ciphertext=ciphertext, secret_key=secret_key, salt=salt), algorithm
    except Exception:
        raise Exception("Incorrect secret key or param")


def load_passwords(
    passwords: Dict[str, str],
    secret_key: str,
    config: dict,
    nested_sep: str = ".",
    sep: str = "$"
) -> Dict[str, Dict[str, str]]:
    """
    Decode + replace variables with password type in original passed dict or indent dict.
    Also returns dict with originally encrypted params

    Args:
        passwords: dict { var route -> encrypted value}
        secret_key:
        config: collection of variables to be changed
        nested_sep: level sep for var route
        sep: fields sep for encrypted strings
    """
    assert nested_sep
    assert sep

    successful_decrypt = {}
    
    for varname, password in passwords.items():

        try:
            decrypted_text, algorithm = decrypt_password(password, secret_key=secret_key, sep=sep)
        except Exception as e:
            print('\t' + e.args[0] + f" for item '{varname}'")
            continue

        # from shared.utils.settings_update import update_data_element
        from ..utils.settings_update import update_data_element
        try:
            config = update_data_element(
                command=varname, arg=decrypted_text,
                data=config, ignore_errors=False, show_values_on_error=False
            )
        except Exception as e:
            print('\t' + e.args[0])
            continue
        
        # save successfully updated param
        successful_decrypt[varname] = {"alg": algorithm, "secret_key": secret_key}

    return successful_decrypt






"""
useful aliases
"""

#region IMPORTS

from typing import (
    Tuple, Sequence,
    Optional, Union, TypeVar, Literal,
    Hashable, Mapping,
    TYPE_CHECKING
)
from typing_extensions import TypeAlias

import os

import logging


#endregion

PathLike: TypeAlias = Union[str, os.PathLike]

T = TypeVar('T')
V = TypeVar('V')




import os
import sys


IS_WINDOWS = sys.platform == 'win32'


def get_cpu_count() -> int:

    if IS_WINDOWS:
        import psutil
        cpu = psutil.cpu_count(logical=False)
    else:
        import subprocess

        try:
            cmd = "lscpu -b -p=Core,Socket | grep -v '^#' | sort -u | wc -l"
            ps = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            output = ps.communicate()[0]
            cpu = int(output)
        except Exception:
            import multiprocessing
            cpu = multiprocessing.cpu_count() // 2

        try:
            cpu = min(cpu, os.sched_getaffinity(0))
        except Exception:
            pass

    return cpu




from typing import Union

import os

# from lockfile import FileLock
from filelock import FileLock

from .aliases import PathLike
from .files import mkdir_of_file


def get_lock(path: PathLike, mode: int = 0o664, timeout: float = -1) -> FileLock:
    """returns lock object with necessary permissions"""

    if not isinstance(path, str):
        path = str(path)

    if not path.endswith('.lock'):
        path += '.lock'

    mkdir_of_file(path)

    return FileLock(path, mode=mode, timeout=timeout)






"""
Special syntax for updating objects through yaml files

TODO add next operations:
    remove (by value)
    pop (by key)
"""

from typing import List, MutableMapping, Any, Union, Dict, Type
from typing_extensions import TypeAlias

import re

#region ALIASES

VARIABLES: TypeAlias = MutableMapping[str, Any]
"""variables dictionary collection"""


VARIABLE_INNER: TypeAlias = Union[VARIABLES, List]
"""object inside variables dictionary"""

#endregion


#region UTILS

def _is_int(s: str) -> bool:
    return bool(re.match(r"^\-?\d+$", s))


def _check_type(value: Any, valid_types: Any, label: str):
    if not isinstance(value, valid_types):
        raise ValueError(f"{label} ({value}) has invalid type for the operation: {type(value)}, requires {valid_types}")


#region getattr/setattr

def _getattr(obj: VARIABLE_INNER, key: Any, default: Any = ...):
    """
    extracts field from the object

    >>> r = dict(a=1, b=2)
    >>> assert _getattr(r, 'a') == 1
    >>> assert _getattr(r, 'c') is ...

    >>> r = [1, 2]
    >>> assert _getattr(r, 1) == 2
    >>> assert _getattr(r, -1) == 2
    >>> assert _getattr(r, 10) is ...

    >>> class R:
    ...     pass
    >>> r = R(); r.a = 1; r.b = 2
    >>> assert _getattr(r, 'a') == 1
    >>> assert _getattr(r, 'c') is ...
    """
    try:
        return obj[key]
    except TypeError:  # not dict-like
        pass
    except (KeyError, IndexError):  # no key index or no index in list
        return default

    return getattr(obj, str(key), default)


def getattr_(obj: VARIABLE_INNER, key: Union[str, int], default: Any = ...):
    """
    extracts field from the object by str/int key

    >>> r = {'1': 1, 2: 2}
    >>> assert getattr_(r, '1') == 1
    >>> assert getattr_(r, 1) == 1
    >>> assert getattr_(r, 2) == 2
    >>> assert getattr_(r, '2') == 2

    >>> r = [1, 2]
    >>> assert getattr_(r, 1) == 2
    >>> assert getattr_(r, -1) == 2
    >>> assert getattr_(r, '-1') == 2

    >>> class R:
    ...     pass
    >>> r = R()
    >>> assert getattr_(r, '1') is ...
    """
    assert key
    if isinstance(key, int) or _is_int(key):  # try integer case
        v = _getattr(obj, int(key), default)
        if v is not default:
            return v
    return _getattr(obj, str(key), default)


def _setattr(obj: VARIABLE_INNER, key: Union[str, int], value: Any):
    """
    sets object attribute
    >>> d = {}
    >>> _setattr(d, 1, 2); _setattr(d, '2', 3); _setattr(d, 'q', 4)
    >>> d
    {1: 2, '2': 3, 'q': 4}

    >>> d = [1, 2]
    >>> _setattr(d, 1, 3); _setattr(d, -2, 2)
    >>> assert d == [2, 3]

    >>> class R:
    ...     pass
    >>> r = R(); _setattr(r, 1, 2)
    >>> assert getattr_(r, 1) == 2
    """
    if isinstance(key, int):
        try:
            obj[key] = value
            return
        except TypeError:
            pass

    key = str(key)
    try:
        obj[key] = value
    except TypeError:
        setattr(obj, key, value)


def setattr_(obj: VARIABLE_INNER, key: Union[str, int], value: Any):
    """
    >>> d = [1, 2]
    >>> setattr_(d, 1, 3); setattr_(d, '-2', 2)
    >>> assert d == [2, 3]
    """
    if isinstance(key, int) or _is_int(key):
        _setattr(obj, int(key), value)
        return
    _setattr(obj, key, value)

#endregion

#endregion


#region OPERATIONS

class AssignOperation:
    """
    base update operation class which performs usual assignment

    >>> d = {'a': [1, 2, 3], 'b': {'c': 'word'}}
    >>> def _(route: str, value: Any):
    ...     AssignOperation.update_variable(route, value, d)
    >>> def err(route: str, value: Any):
    ...     try:
    ...         _(route, value)
    ...     except ValueError:
    ...         pass
    >>> _('d', 1)
    >>> _('a.0', 0)
    >>> _('a.-1', '1')
    >>> _('b.c', 'new')
    >>> _('b.d', 'd')
    >>> err('b.c.d', 0)
    >>> err('x.y', 0)
    >>> _('x', {})
    >>> _('x.y', 0)
    >>> d
    {'a': [0, 2, '1'], 'b': {'c': 'new', 'd': 'd'}, 'd': 1, 'x': {'y': 0}}
    """

    label = ''

    var_required = False
    """whether the changing variable itself must be present"""

    @classmethod
    def get_view(cls, route: str, value: Any, show_value: bool = True) -> str:
        label = f"@{cls.label}" if cls.label else ''
        if not show_value:
            value = '***'
        return f"{route}{label}: {value}"

    @classmethod
    def update_value(cls, field_value: Any, value: Any):
        """
        update current field according to value (and operation of class)
        this method must be overridden in subclasses

        Args:
            field_value: field to be changed/updated
            value: value to perform updates

        Returns:
            Ellipsis if no field update is required otherwise some new field value
        """
        return value

    @classmethod
    def update_variable(
        cls,
        route: str,
        value: Any,
        variables: VARIABLES,
        level_sep: str = '.',
        show_value_on_err: bool = True
    ):

        parts = route.split(level_sep)
        if not parts or not all(parts):
            raise ValueError(f"unknown variable route: {route} ({cls.get_view(route, value, show_value_on_err)})")

        _variables = variables
        done = []
        while True:
            p, *parts = parts
            var = getattr_(_variables, p)
            if var is Ellipsis:  # not found
                if parts or cls.var_required:  # if there are levels further it is the last level but presence is required
                    raise ValueError(f"{p} variable not found ({cls.get_view(route, value, show_value_on_err)})")

            done.append(p)

            if not parts:  # latest level reached
                try:
                    v = cls.update_value(var, value)
                except ValueError as e:
                    message = f"{level_sep.join(done)}: {e.args[0]} ({cls.get_view(route, value, show_value_on_err)})"
                    raise ValueError(message)
                if v is not Ellipsis:
                    setattr_(_variables, p, v)
                return

            _variables = var
            if isinstance(_variables, (bool, int, float, str, bytes)):  # simple type
                raise ValueError(
                    f"{route} field not found ({cls.get_view(route, value, show_value_on_err)}) "
                    f"because {level_sep.join(done)} is {type(_variables)} (cannot go deeper)"
                )


class SetDefaultOperation(AssignOperation):
    """
    >>> d = {'a': [1, 2, 3], 'b': {'c': 'word'}}
    >>> def _(route: str, value: Any):
    ...     SetDefaultOperation.update_variable(route, value, d)
    >>> def err(route: str, value: Any):
    ...     try:
    ...         _(route, value)
    ...     except ValueError:
    ...         pass
    >>> _('d', 1)
    >>> _('a.0', 0)
    >>> _('a.-1', '1')
    >>> _('b.c', 'new')
    >>> _('b.d', 'd')
    >>> err('b.c.d', 0)
    >>> err('x.y', 0)
    >>> _('x', {})
    >>> _('x.y', 0)
    >>> d
    {'a': [1, 2, 3], 'b': {'c': 'word', 'd': 'd'}, 'd': 1, 'x': {'y': 0}}
    """
    label = 'setdefault'
    var_required = False

    @classmethod
    def update_value(cls, field_value: Any, value: Any):
        if field_value is Ellipsis:
            return value
        return field_value


class BindOperation(AssignOperation):
    """
    >>> _list = [1, 2, 3]
    >>> _dict = {'c': 'word'}
    >>> d = {'a': _list, 'b': _dict, 'c': _list}
    >>> def _(route: str, value: Any):
    ...     BindOperation.update_variable(route, value, d)
    >>> def err(route: str, value: Any):
    ...     try:
    ...         _(route, value)
    ...     except ValueError:
    ...         pass
    >>> _('a', [1, 2])
    >>> _('c', [-1, -2])
    >>> err('b.e', [1])
    >>> err('b.c', [2])
    >>> _('b', {'new': 1})
    >>> d
    {'a': [-1, -2], 'b': {'new': 1}, 'c': [-1, -2]}
    >>> assert d['c'] is _list
    >>> assert d['b'] is _dict
    """
    label = 'bind'

    @classmethod
    def update_value(cls, field_value: Any, value: Any):
        if field_value is Ellipsis:
            if cls.var_required:
                raise ValueError(f'no var {field_value}')
            else:
                _check_type(value, (list, dict), 'value')
                return value

        _check_type(field_value, (list, dict), 'field')
        _check_type(value, (list, dict), 'value')

        field_value.clear()
        if isinstance(field_value, list):
            field_value.extend(value)
        else:
            field_value.update(value)
        return field_value


class AppendOperation(AssignOperation):
    """
    >>> d = {'a': [1, 2, 3], 'b': {'c': []}}
    >>> def _(route: str, value: Any):
    ...     AppendOperation.update_variable(route, value, d)
    >>> def err(route: str, value: Any):
    ...     try:
    ...         _(route, value)
    ...     except ValueError:
    ...         pass
    >>> _('a', 4)
    >>> _('b.c', 5)
    >>> err('b', 1)
    >>> err('b.d', 1)
    >>> d
    {'a': [1, 2, 3, 4], 'b': {'c': [5]}}
    """

    label = 'append'
    var_required = True
    default_value_type = list

    @classmethod
    def _update_value_exactly(cls, field_value: list, value: Any) -> list:
        field_value.append(value)
        return field_value

    @classmethod
    def _update_value(cls, field_value: list, value: Any) -> list:
        if field_value is Ellipsis:
            if cls.var_required:
                raise ValueError("field value is not set")
            else:
                field_value = cls.default_value_type()

        _check_type(field_value, cls.default_value_type, 'field')
        return cls._update_value_exactly(field_value, value)

    @classmethod
    def update_value(cls, field_value: list, value: Any) -> list:
        # _check_type(value, list, 'value')
        return cls._update_value(field_value, value)


class TryAppendOperation(AppendOperation):
    """
    >>> d = {'a': [1, 2, 3], 'b': {'c': []}}
    >>> def _(route: str, value: Any):
    ...     TryAppendOperation.update_variable(route, value, d)
    >>> _('a', 4)
    >>> _('b.c', 5)
    >>> _('b.d', 1)
    >>> d
    {'a': [1, 2, 3, 4], 'b': {'c': [5], 'd': [1]}}
    """
    label = 'try-append'
    var_required = False


class ExtendOperation(AppendOperation):
    """
    >>> d = {'a': [1, 2, 3], 'b': {'c': []}}
    >>> def _(route: str, value: Any):
    ...     ExtendOperation.update_variable(route, value, d)
    >>> def err(route: str, value: Any):
    ...     try:
    ...         _(route, value)
    ...     except ValueError:
    ...         pass
    >>> _('a', [4])
    >>> err('b.c', 5)
    >>> _('b.c', [5, 6, 7])
    >>> err('b', 1)
    >>> err('b.d', 1)
    >>> d
    {'a': [1, 2, 3, 4], 'b': {'c': [5, 6, 7]}}
    """

    label = 'extend'
    var_required = True

    @classmethod
    def _update_value_exactly(cls, field_value: list, value: list) -> list:
        field_value.extend(value)
        return field_value

    @classmethod
    def update_value(cls, field_value: list, value: list) -> list:
        _check_type(value, list, 'value')
        return cls._update_value(field_value, value)


class TryExtendOperation(ExtendOperation):
    """
    >>> d = {'a': [1, 2, 3], 'b': {'c': []}}
    >>> def _(route: str, value: Any):
    ...     TryExtendOperation.update_variable(route, value, d)
    >>> def err(route: str, value: Any):
    ...     try:
    ...         _(route, value)
    ...     except ValueError:
    ...         pass
    >>> _('a', [4])
    >>> err('b.c', 5)
    >>> _('b.c', [5, 6, 7])
    >>> err('b', 1)
    >>> err('b.d', [1, 2])
    >>> d
    {'a': [1, 2, 3, 4], 'b': {'c': [5, 6, 7], 'd': [1, 2]}}
    """

    label = 'try-extend'
    var_required = False


class UpdateOperation(AppendOperation):
    """
    >>> d = {'a': [1, 2, 3], 'b': {'c': {}}}
    >>> def _(route: str, value: Any):
    ...     UpdateOperation.update_variable(route, value, d)
    >>> def err(route: str, value: Any):
    ...     try:
    ...         _(route, value)
    ...     except ValueError:
    ...         pass
    >>> err('a', [4])
    >>> err('b.c', 5)
    >>> _('b.c', {5: 6, 7: 8})
    >>> err('b', 1)
    >>> err('b.d', 1)
    >>> d
    {'a': [1, 2, 3], 'b': {'c': {5: 6, 7: 8}}}
    """

    label = 'update'
    var_required = True
    default_value_type = dict

    @classmethod
    def _update_value_exactly(cls, field_value: dict, value: dict) -> dict:
        field_value.update(value)
        return field_value

    @classmethod
    def update_value(cls, field_value: dict, value: dict) -> list:
        _check_type(value, dict, 'value')
        return cls._update_value(field_value, value)


class TryUpdateOperation(UpdateOperation):
    """
    >>> d = {'a': [1, 2, 3], 'b': {'c': {}}}
    >>> def _(route: str, value: Any):
    ...     TryUpdateOperation.update_variable(route, value, d)
    >>> def err(route: str, value: Any):
    ...     try:
    ...         _(route, value)
    ...     except ValueError:
    ...         pass
    >>> err('a', [4])
    >>> err('b.c', 5)
    >>> _('b.c', {5: 6, 7: 8})
    >>> err('b', 1)
    >>> _('b.d', {1: 2})
    >>> d
    {'a': [1, 2, 3], 'b': {'c': {5: 6, 7: 8}, 'd': {1: 2}}}
    """

    label = 'try-update'
    var_required = False


#endregion


label_to_operation: Dict[str, Type[AssignOperation]] = {
    c.label: c for c in (
        AssignOperation, SetDefaultOperation, BindOperation,
        AppendOperation, TryAppendOperation,
        ExtendOperation, TryExtendOperation,
        UpdateOperation, TryUpdateOperation
    )
}


def update_data_element(
    command: str,
    arg: Any,
    data: VARIABLES,
    show_values_on_error: bool = True,
    ignore_errors: bool = False,
) -> VARIABLES:
    """
    performs update command on existing data
    Args:
        command: key@operation string
        arg: command argument
        data: data to be updated
        show_values_on_error: whether to show values in error messages
        ignore_errors: whether to skip errors

    Returns:
        updated data

    Notes:
        updates data inplace
    """
    if '@' in command:  # extract command
        try:
            key, label = command.rsplit('@', maxsplit=1)
            assert label in label_to_operation, (label, label_to_operation.keys())
            assert key, command
            cls = label_to_operation[label]
        except Exception:
            if ignore_errors:
                return data
            raise
    else:  # assign
        key = command
        cls = AssignOperation

    try:
        cls.update_variable(
            route=key, value=arg,
            variables=data, show_value_on_err=show_values_on_error
        )
    except Exception:
        if ignore_errors:
            return data
        raise

    return data


def update_data(
    commands: Dict[str, Any],
    data: VARIABLES,
    show_values_on_error: bool = True,
    ignore_errors: bool = False
) -> VARIABLES:
    """
    performs update commands on existing data
    Args:
        commands: dict { key@operation -> value }
        data: data to be updated
        show_values_on_error: whether to show values in error messages
        ignore_errors: whether to skip errors

    Returns:
        updated data

    Notes:
        updates data inplace

    >>> class R: pass
    >>> r = R()
    >>> _list = [5, 6, 7]
    >>> r.x = {'a': 1, 'b': {'c': [1, 2], 'd': _list}}
    >>> r.y = 11
    >>> r.z = [0, 1, 2]
    >>> r.lst = _list
    >>> cmds = {
    ...     'u': 'u',
    ...     'y@setdefault': 0,
    ...     'lst@bind': [7, 6, 5],
    ...     'x.c@bind': [3],
    ...     'z@append': 4,
    ...     'z@extend': [5, 6],
    ...     'v@try-append': [-1],
    ...     'n@try-extend': [-1],
    ...     'x.b@update': {'new': 'key'}
    ... }
    >>> _ = update_data(cmds, r)
    >>> r.__dict__
    {'x': {'a': 1, 'b': {'c': [1, 2], 'd': [7, 6, 5], 'new': 'key'}, 'c': [3]}, 'y': 11, 'z': [0, 1, 2, 4, 5, 6], 'lst': [7, 6, 5], 'u': 'u', 'v': [[-1]], 'n': [-1]}
    """

    if not commands:
        return data

    for command, arg in commands.items():
        data = update_data_element(
            command, arg,
            data=data,
            show_values_on_error=show_values_on_error, ignore_errors=ignore_errors
        )

    return data









from typing import MutableMapping, Mapping, Any, Optional, Dict

import os
import io
from pathlib import Path

from .docrender_utils.aliases import PathLike
from .docrender_utils.files import read_yaml, is_like_relative_path, read_text
from .docrender_utils.lock import get_lock


def load_settings_from_yaml(
    path: PathLike,
    object_to_update: MutableMapping[str, Any],
    base_dir: Optional[PathLike] = None,
    show_update_errors: bool = True,
):
    """
    performs settings update from yaml file with extra actions such as
        `varsubst`, relative paths resolution, yaml commands execution

    Args:
        path: path to yaml file
        object_to_update: object to be updated using this file, usually locals()
        base_dir: base directory to translate relative paths; None means to disable this translation
        show_update_errors: whether to verbose updating errors

    """
    try:
        text = read_text(path)
        try:
            from varsubst import varsubst
            text = varsubst(text)
        except ImportError:
            pass

        new_vars = read_yaml(io.StringIO(text))
        if new_vars:
            if base_dir:
                new_vars = {
                    k: (
                        str(Path(base_dir, v).resolve().absolute())
                        if isinstance(v, str) and is_like_relative_path(v)
                        else v
                    )
                    for k, v in new_vars.items()
                }

            # much extended version of object_to_update.update(_new_vars)
            from .settings_update import update_data
            update_data(
                commands=new_vars, data=object_to_update,
                show_values_on_error=show_update_errors,
            )

    except Exception as e:
        print(f"Can't load yaml settings ({str(path)})", e)


def load_passwords_from_yaml(
    passwords_file: PathLike,
    secret_key_file: PathLike,
    object_to_update: MutableMapping[str, Any],
    password_sep: str = '$'
) -> Dict[str, Dict[str, str]]:
    """
    performs settings update using passwords file and secret
    Args:
        passwords_file: path to yaml with encrypted variables/commands
        secret_key_file: path to secret file
        object_to_update: object to be updated using this file, usually locals()
        password_sep:

    Returns:

    """
    with get_lock(passwords_file, timeout=10):
        passwords_file = Path(passwords_file)
        if not passwords_file.exists():
            # print(f"Not found passwords file: {passwords_file}")
            return {}

        passwords = read_yaml(passwords_file)
        if not passwords:
            return {}
        
    from ..encryption.utils import load_secret
    secret_key = load_secret(secret_key_file)
    if not secret_key:
        return {}

    from ..encryption.encryptor import load_passwords

    try:
        # print(f"Reading passwords file: {passwords_file}")
        return load_passwords(
            passwords=passwords,
            config=object_to_update,
            secret_key=secret_key,
            sep=password_sep,
        )
    except Exception:
        from traceback import print_exc
        print_exc()
        return {}


def _decrypt_vars(
    dct: Dict[str, str],
    secret_key_file: Optional[PathLike] = None,
    password_sep: str = '$',
) -> Optional[Dict[str, str]]:
    if not dct:
        return
    if not secret_key_file:
        print('secret key file is not set but is required to parse some env variables')
        return

    from ..encryption.utils import load_secret
    secret_key = load_secret(secret_key_file)
    if not secret_key:
        return

    from ..encryption.encryptor import decrypt_password
    res = {}
    for k, v in dct.items():
        try:
            v = decrypt_password(v, secret_key=secret_key, sep=password_sep)[0]
            res[k] = v
        except Exception as e:
            print('\t' + e.args[0])

    return res


def load_vars_from_env(
    object_to_update: MutableMapping[str, Any],
    secret_key_file: Optional[PathLike] = None,
    password_sep: str = '$',
    prefix: str = 'TRANSLATE_'
):
    """
    updates object according to env variables rules with encrypted data allowed
    Args:
        object_to_update:
        secret_key_file:
        password_sep:
        prefix: prefix to select env variables to be used

    """

    # locals().update(
    #     parse_vars(
    #         prefix='DREAMOCR_',
    #         initial_vars=locals(),
    #         dict_level_separator='__'
    #     )
    # )

    source_vars = dict(os.environ)
    # filter by prefix
    source_vars = {
        k[len(prefix):]: v for k, v in source_vars.items()
        if k.startswith(prefix)
    }
    if not source_vars:
        return

    # suffix = '_ENCRYPTED'
    # encrypted_vars = {
    #     k[:-len(suffix)]: v for k, v in source_vars.items() if k.endswith(suffix)
    # }
    from ..encryption.encryptor import is_like_secret
    encrypted_vars = {
        k: v for k, v in source_vars.items() if is_like_secret(v)
    }
    if encrypted_vars:
        # remove all encrypted variables from source
        source_vars = {k: v for k, v in source_vars.items() if k not in encrypted_vars}

        print(f"Decrypt vars from environment...")
        decrypted_vars = _decrypt_vars(encrypted_vars, secret_key_file=secret_key_file, password_sep=password_sep)
        if decrypted_vars:
            source_vars.update(decrypted_vars)

    if source_vars:
        from env2dict import parse_vars
        object_to_update.update(
            parse_vars(
                prefix='',
                source=source_vars,
                initial_vars=object_to_update,
                dict_level_separator='__'
            )
        )


def save_settings_debug(settings: Mapping[str, Any]):
    """
    save all settings for debug only purposes
    """
    _s = settings.get("SAVE_SETTINGS_DEBUG", None)
    if _s:
        import json
        import inspect

        _d = {
            k: v
            for k, v in settings.items()
            if not (
                    k.startswith("__")
                    or len(k.strip("_")) < 2
                    or callable(v)
                    or inspect.ismodule(v)
                    or inspect.isclass(v)
            )
        }
        _d["environment"] = dict(os.environ.items())

        p = Path(_s)
        p.parent.mkdir(exist_ok=True, parents=True)
        p.write_text(
            json.dumps(
                _d, default=lambda o: "<not serializable>", indent=2, sort_keys=True
            ),
            encoding="utf-8",
        )




from typing import Optional
from abc import ABC, abstractmethod


class BaseEncryptor(ABC):

    label = ''
    """this algorithm label, using in full encryption"""
    
    @classmethod
    @abstractmethod
    def generate_salt(cls) -> str:
        """Return storaged salt"""
        
    @classmethod
    @abstractmethod
    def generate_secret(cls) -> str:
        """Return generated secret key in base64"""
    
    @classmethod
    @abstractmethod
    def encrypt(cls, plain_text: str, secret_key: str, salt: str) -> str:
        """Return encrypted string"""
        
        
    @classmethod
    @abstractmethod
    def decrypt(cls, ciphertext: str, secret_key: str, salt: str) -> str:
        """Return decrypted string"""

    @classmethod
    def encrypt_full(cls, plain_text: str, secret_key: str, salt: Optional[str] = None, sep: str = '$') -> str:
        """
        encrypts plain text to full secret cipher with format ALGO$SALT$CIPHER
        Args:
            plain_text:
            secret_key:
            salt:
            sep:

        Returns:

        """

        salt = salt or cls.generate_salt()

        try:
            ciphertext = cls.encrypt(plain_text=plain_text, secret_key=secret_key, salt=salt)
        except Exception:
            raise ValueError("Invalid secret")

        return sep.join(
            (cls.label.upper(), salt, ciphertext)
        )





#region IMPORTS

from typing import Sequence, Literal, Optional, Dict, Any, Union, List, Iterable, Tuple

import os
import io
import shutil
from pathlib import Path
import contextlib

from .aliases import PathLike
from .system import IS_WINDOWS

#endregion


@contextlib.contextmanager
def umask_context(umask: int):
    """sets umask on operation and restores after"""
    old_umask = os.umask(umask)
    try:
        yield
    finally:
        os.umask(old_umask)


#region RM/TOUCH

def _mkdir(path: Path):
    with umask_context(0o002):
        path.mkdir(parents=True, exist_ok=True)


def mkparents(path: PathLike):
    """equals to mkdir -p $(dirname path)"""
    _mkdir(Path(path).parent)


def mkdir_of_file(file_path: PathLike):
    """
    для этого файла создаёт папку, в которой он должен лежать
    """
    mkparents(file_path)


def mkdir(path: PathLike):
    """mkdir with parents"""
    _mkdir(Path(path))


def touch(path: PathLike):
    """makes empty file, makes directories for this file automatically"""
    mkdir_of_file(path)
    Path(path).touch()


def rmdir(path: PathLike):
    """rm dir without errors"""
    shutil.rmtree(path, ignore_errors=True)


def copy_file(source: PathLike, dest: PathLike):
    """performs file copying with target directory auto creation"""
    if os.path.exists(dest) and os.path.samefile(source, dest):
        return
    mkdir_of_file(dest)
    shutil.copyfile(source, dest)


def move_file(source: PathLike, dest: PathLike):
    """performs file moving with target directory auto creation"""
    mkdir_of_file(dest)
    shutil.move(source, dest)


#endregion

#region READ/WRITE/PARSE/DUMP


def read_json(path: PathLike) -> Union[Dict[str, Any], List]:
    import orjson
    if str(path).endswith('.gz'):
        import gzip
        with gzip.GzipFile(str(path), 'rb') as f:
            r = f.read()
    else:
        r = Path(path).read_bytes()
    return orjson.loads(r)


def to_json_bytes(data: Union[Dict, List, Any]) -> bytes:
    import orjson
    return orjson.dumps(data, option=orjson.OPT_INDENT_2 | orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_NON_STR_KEYS)


def save_json(path: PathLike, data: Union[Dict, List, Any]):
    mkdir_of_file(path)
    Path(path).write_bytes(
        to_json_bytes(data)
    )

    # dump = json.dumps(
    #     data.asdict() if hasattr(data, 'asdict') else data,
    #     ensure_ascii=False,
    #     indent=2
    # )
    # result_tmp = str(path)
    # Path(result_tmp).write_bytes(dump.encode('utf-8', 'ignore'))


def write_json(path: PathLike, data: Union[Dict, List, Any]):
    save_json(path, data)


def write_text(result_path: PathLike, text: str, encoding: str = 'utf-8'):
    mkdir_of_file(result_path)
    result_tmp = str(result_path) + '.tmp'
    Path(result_tmp).write_text(text, encoding=encoding, errors='ignore')
    if os.path.exists(result_path):
        os.remove(result_path)
    Path(result_tmp).rename(result_path)


def read_text(result_path: PathLike, encoding: str = 'utf-8'):
    return Path(result_path).read_text(encoding=encoding, errors='ignore')


def write_pickle(result_path: PathLike, data: Any):
    mkdir_of_file(result_path)
    import pickle
    dump = pickle.dumps(data)  # , protocol=4)
    result_tmp = str(result_path) + '.tmp'
    Path(result_tmp).write_bytes(dump)
    if os.path.exists(result_path):
        os.remove(result_path)
    Path(result_tmp).rename(result_path)


def read_pickle(result_path: PathLike):
    import pickle
    data = pickle.loads(Path(result_path).read_bytes())
    return data


def dumps_yaml(dct: Dict[str, Any]):
    """
    converts dictionary to yaml format string

    >>> print(dumps_yaml(dict(a=1, b=2, c='abc', d={'a': [1, 2], 'b': (1, 2), 'c': {1: 1, 2: 2}})).strip())
    a: 1
    b: 2
    c: abc
    d:
      a:
      - 1
      - 2
      b: !!python/tuple
      - 1
      - 2
      c:
        1: 1
        2: 2
    """
    from ._yaml import yaml_dump
    return yaml_dump(dct)


def write_yaml(path: PathLike, data: Dict):
    """writes dict to yaml"""
    write_text(path, dumps_yaml(data))


def parse_yaml(
    source: Union[PathLike, io.StringIO]
) -> Dict[str, Any]:
    """
    reads dict from .yaml file or string io

    >>> dct = {'a': 1, 'b': [1, 2, 3], 'c': {0: 0, 1: 1}}
    >>> s = dumps_yaml(dct)
    >>> assert parse_yaml(io.StringIO(s)) == dct
    """
    from ._yaml import yaml_load
    if isinstance(source, (str, Path)):
        with open(source, 'r', encoding='utf-8') as f:
            file_configs = yaml_load(f)
    else:
        file_configs = yaml_load(source)
    return file_configs


def read_yaml(
    source: Union[PathLike, io.StringIO]
) -> Dict[str, Any]:
    return parse_yaml(source)

#endregion


#region LOOKS LIKE

def looks_like(file: PathLike, pattern: Union[str, Tuple[str, ...]]):
    return (
        file if isinstance(file, str) else str(file)
    ).lower().endswith(pattern)


def looks_like_pdf(file: PathLike) -> bool:
    """
    >>> looks_like_pdf('p')
    False
    >>> looks_like_pdf('p.pdf')
    True
    >>> looks_like_pdf('p.PDF')
    True

    """
    return looks_like(file, '.pdf')


def looks_like_json(file: PathLike):
    return looks_like(file, '.json')


def looks_like_doc(file: PathLike):
    return looks_like(file, ('.doc', '.docx', '.rtf'))


def looks_like_text(file: PathLike):
    return looks_like(file, ('.txt', '.text'))


def looks_like_tiff(file: PathLike):
    return looks_like(file, ('.tif', '.tiff'))


def looks_like_png(file: PathLike):
    return looks_like(file, '.png')


def looks_like_jpeg(file: PathLike):
    return looks_like(file, ('.jpg', '.jpeg'))


def is_pdf_file(fpath: PathLike):
    return looks_like_pdf(fpath) and os.path.isfile(fpath)


def is_doc_file(fpath: PathLike):
    return looks_like_doc(fpath) and os.path.isfile(fpath)


def is_json_file(fpath: PathLike):
    return looks_like_json(fpath) and os.path.isfile(fpath)


def is_tiff_file(fpath: PathLike):
    return looks_like_tiff(fpath) and os.path.exists(fpath)


def get_first_file_endswith(
    files: Iterable[PathLike],
    endswith: Union[str, Sequence[str]]
) -> Optional[PathLike]:
    """
    searches for the first file with required end (like extension but not mandatory)
    Args:
        files: sequence of files candidates (usually files in the directory)
        endswith: pattern for the file end (or sequence of valid patterns)

    Returns:
        found file or None if not found
    """
    if not isinstance(endswith, str):
        endswith = tuple(endswith)
    return next(
        (p for p in files if Path(p).name.endswith(endswith)),
        None
    )


#endregion

def isfile_casesensitive(path: PathLike) -> bool:
    """
    checks whether the file exists exactly with this name

    Notes:
        on Windows test.txt TEST.txt are same files for many operations
    """
    if not os.path.isfile(path):
        return False
    directory, filename = os.path.split(path)
    return filename in os.listdir(directory)

#region PATHS


def doc_name(fpath: PathLike):
    """
    само имя файла, без пути и без расширения
    """
    return Path(fpath).stem


def full_path(file: PathLike):
    """resolved abspath"""
    return str(Path(file).absolute().resolve())


_illegal_path_symbols = set('№#$%&+;<=>?^`{}')


def is_like_relative_path(string: str) -> bool:
    """
    checks if string is like relative path (but bot strict)

    >>> is_like_relative_path('asferefefwfw11221')
    False
    >>> is_like_relative_path('./dererew')
    True
    >>> is_like_relative_path('models/model1')
    True
    >>> is_like_relative_path('C:\\\\re')
    False
    """
    if any(s in _illegal_path_symbols for s in string):
        return False

    is_path = '/' in string
    if not is_path and IS_WINDOWS:
        is_path = '\\' in string

    if not is_path:
        return False

    if IS_WINDOWS:
        return not string[1] == ':'
    return not string.startswith('/') and ':' not in string


def resolve_relative(
    relative_path: str,
    paths_to_start: Optional[Sequence[str]] = None
):
    """
    resolves abs or relative path

    Args:
        relative_path: path to resolve
        paths_to_start: additional dirs to try to resolve

    Returns:
        resolved path
    """
    # check existence; if relative -- check if relative from curdir
    if not os.path.exists(relative_path):
        if not is_like_relative_path(relative_path):  # absolute path case
            raise FileNotFoundError(f"file on absolute path '{relative_path}' is not exists")

        if not paths_to_start:
            paths_to_start = []
        else:
            paths_to_start = list(paths_to_start)

        for start in paths_to_start:
            assert os.path.isdir(start), f"argument path {start} is not an existing dir"

            path_new = str(Path(start).resolve().joinpath(relative_path).resolve())
            if os.path.exists(path_new):
                resolved = path_new
                break
        else:
            raise FileNotFoundError(
                f"cannot resolve path '{relative_path}' for any CWD from {tuple([os.getcwd()] + paths_to_start)}"
            )

        print(f"path {relative_path} is resolved as {resolved}")
        return resolved

    return os.path.abspath(relative_path)


def set_unix_sep(path: PathLike) -> str:
    """replaces backslash with unix /"""
    return str(path).replace(os.sep, '/')


def get_relative_path(path: PathLike, relative_to: PathLike) -> str:
    """
    returns relative path as string in unix-style format

    >>> assert get_relative_path('./a/b/c', './a') == 'b/c'
    >>> assert get_relative_path('./a', './a/b/c') == '../..'
    """
    return set_unix_sep(os.path.relpath(path, relative_to))


def isin(path: PathLike, parent: PathLike) -> bool:
    """
    checks whether the path is in other path
    Args:
        path: path to file or directory
        parent: the path to check whether it is its parent

    Returns:

    >>> assert isin('a/b/c', 'a/b')
    >>> assert isin('a/b/c', 'a')
    >>> assert not isin('a/b/c', 't/b')
    """
    return not get_relative_path(path, parent).startswith('..')

#endregion

#region META

_scales = {
    'b': 1,
    'kb': 1024,
    'mb': 1024 * 1024,
    'gb': 1024 * 1024 * 1024
}


def file_size(file: PathLike, scale: Literal['b', 'kb', 'mb', 'gb'] = 'b'):
    """file size"""
    return Path(file).stat().st_size / _scales[scale]


def check_file_is_open(fn: PathLike, prefix: str = ' '):
    import psutil
    p = psutil.Process()
    for f in p.open_files():
        # print(f)
        if f.path == fn:
            print('!', prefix, "File is open:", fn)
            return
    print('!', prefix, "File is closed:", fn)


def assert_files_existence(files: Union[PathLike, Sequence[PathLike]], nonempty: bool = True):
    """asserts whether all input files exist"""
    if not isinstance(files, (list, tuple)):
        files = [files]

    for use, func, message in (
        (True, os.path.isfile, "next files don't exist or are not files"),
        (nonempty, os.path.getsize, "next files exist but are empty")
    ):
        if not use:
            continue

        mask = [func(f) for f in files]
        if not all(mask):
            raise FileNotFoundError(
                f"{message} ({sum(mask)}/{len(mask)}):\n\t" +
                '\n\t'.join(str(p) for f, p in zip(mask, files) if not f)
            )

#endregion


def concat_text_files(inpaths: Sequence[PathLike], outpath: PathLike):
    mkdir_of_file(outpath)
    with open(outpath, 'wb') as wfd:
        for f in inpaths:
            with open(f, 'rb') as fd:
                shutil.copyfileobj(fd, wfd)
                wfd.write(b"\n")


def nl(
    text: Union[str, os.PathLike, Iterable[str]],
    starts: int = 1,
    delimiter: str = '\t',
    only_nonempty: bool = True,
    left_align: bool = False,
    filler: str = ' ',
    save_to: Optional[PathLike] = None,
) -> List[str]:
    """
    add lines numbers at the start of each string

    Args:
        text: string sequence or a path to file
        starts: the number of first string
        delimiter: the text to put between numbers and original strings
        only_nonempty: numerate only nonempty lines
        left_align: align numbers at left column part, not right
        filler: character to fill empty space in numbers
        save_to: file to save the result

    Returns:
        list of new lines

    >>> def _(*args, **kwargs):
    ...     _lines = nl(*args, **kwargs)
    ...     print(f'{chr(10)}'.join(_lines))  # 10 is \n

    >>> t = ['line 1', '  line 2', '', '', 'line 5']
    >>> _(t, delimiter='   ')
    1   line 1
    2     line 2
    <BLANKLINE>
    <BLANKLINE>
    3   line 5
    >>> _(t, only_nonempty=False, delimiter=' --', starts=0)
    0 --line 1
    1 --  line 2
    2 --
    3 --
    4 --line 5
    >>> t = list('123456789abcd')
    >>> _(t, only_nonempty=False, delimiter='<-->')
     1<-->1
     2<-->2
     3<-->3
     4<-->4
     5<-->5
     6<-->6
     7<-->7
     8<-->8
     9<-->9
    10<-->a
    11<-->b
    12<-->c
    13<-->d
    >>> _(t, only_nonempty=False, delimiter='<-->', filler='0')
    01<-->1
    02<-->2
    03<-->3
    04<-->4
    05<-->5
    06<-->6
    07<-->7
    08<-->8
    09<-->9
    10<-->a
    11<-->b
    12<-->c
    13<-->d
    >>> _(t, only_nonempty=False, delimiter='<-->', left_align=True)
    1 <-->1
    2 <-->2
    3 <-->3
    4 <-->4
    5 <-->5
    6 <-->6
    7 <-->7
    8 <-->8
    9 <-->9
    10<-->a
    11<-->b
    12<-->c
    13<-->d
    """

    assert filler

    if isinstance(text, (str, Path)):
        text = read_text(text).split('\n')

    lines: List[str] = text

    lines_count = (
        sum(bool(line.strip()) for line in lines)
        if only_nonempty
        else len(lines)
    )

    max_number_len = len(str(lines_count + starts))

    empty = ' ' * (max_number_len + len(delimiter))

    def _get_text(number: int, line: str):
        if only_nonempty and not line.strip():
            return empty

        n = str(number)
        if left_align:
            n = n.ljust(max_number_len, filler)
        else:
            n = n.rjust(max_number_len, filler)

        return n

    # result = [
    #     f"{_get_text(i, line)}{delimiter}{line}"
    #     for i, line in enumerate(lines, starts)
    # ]

    result = []
    k = starts
    for line in lines:
        if only_nonempty and not line.strip():
            result.append(empty + line)
        else:
            s = _get_text(k, line)
            k += 1
            result.append(
                s + delimiter + line
            )

    if save_to:
        mkdir_of_file(save_to)
        write_text(save_to, '\n'.join(result))

    return result











from typing import Optional

from pathlib import Path

from ..utils.docrender_utils.aliases import PathLike, T, V
from ..utils.docrender_utils.files import read_text

from .encryptor import decrypt_password, is_like_secret


def load_secret(path: PathLike, ignore_errors: bool = True) -> Optional[str]:
    """returns secret key or None on problems"""
    secret_key_file = Path(path)
    if not secret_key_file.exists():
        m = "Secret key file not found but is required!"
        if ignore_errors:
            print(m)
            return
        raise Exception(m)

    secret_key = read_text(secret_key_file)
    if not secret_key.strip():
        m = "Empty secret key!"
        if ignore_errors:
            print(m)
            return
        raise Exception(m)

    return secret_key


def decrypt_passwords(
    container: T,
    secret_key: str = '',
    secret_key_file: Optional[PathLike] = None,
    sep: str = "$"
) -> T:
    """
    decrypts encrypted string values in json-like object
    Args:
        container: json-like object or simple type instance
        secret_key:
        secret_key_file: path to secret key
        sep:

    Returns:
        same object with decrypted values

    Notes:
        - any and only one from secret_key/secret_key_file must be specified
        - will not change original object
        - for not json-like object will return itself without raising errors and performing transformations
        - will not check/load secret key if there is no real need
    """

    assert bool(secret_key) + bool(secret_key_file) == 1, 'secret_key or secret_key_file must be specified but not both'

    def decrypt(value: str) -> str:
        nonlocal secret_key
        if not secret_key:  # load key
            secret_key = load_secret(secret_key_file, ignore_errors=False)

        return decrypt_password(value, secret_key=secret_key, sep=sep)[0]

    def run(obj: V) -> V:
        if isinstance(obj, (list, tuple)):
            return [run(i) for i in obj]

        if isinstance(obj, dict):
            return {k: run(v) for k, v in obj.items()}

        if isinstance(obj, str) and obj and is_like_secret(obj):
            return decrypt(obj)

        return obj

    return run(container)






"""
creates yaml-based functions according to existing machine configuration state
"""

from typing import Dict

import yaml

try:
    from yaml import CLoader as Loader, CDumper as Dumper
    LIBYAML_AVAILABLE = True
except ImportError:
    from yaml import Loader, Dumper
    LIBYAML_AVAILABLE = False

# to ignore &id001 like links
Dumper.ignore_aliases = lambda *args: True

_dump_kwargs = dict(
    default_flow_style=False
)


def yaml_load(stream) -> Dict:
    return yaml.load(stream, Loader=Loader)


def yaml_dump(data, **kwargs) -> str:
    return yaml.dump(
        data, Dumper=Dumper, encoding='utf-8', allow_unicode=True, **{**_dump_kwargs, **kwargs}
    ).decode()




from pathlib import Path

import setuptools


def parse_requirements(requirements: str):
    with open(requirements) as f:
        return [
            l.strip('\n') for l in f if l.strip('\n') and not l.startswith('#')
        ]


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="conf-loaders", 
    version=Path('version.txt').read_text(encoding='utf-8').strip(),
    author="Demetry Pascal",
    author_email="qtckpuhdsa@gmail.com",
    maintainer='Demetry Pascal',
    description="configuration loading utils",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PasaOpasen/conf-loaders",
    keywords=['configuration', 'conf'],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=parse_requirements('./requirements.txt'),
    
)



import base64
from os import urandom

from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from .base import BaseEncryptor


class Encryptor(BaseEncryptor):

    label = 'aes'
    
    @classmethod
    def generate_salt(cls) -> str:
        """Generate salt"""
        random_bytes: bytes = urandom(16)
        return base64.b64encode(random_bytes).decode("utf-8")

    @classmethod
    def generate_secret(cls) -> str:
        """Return generated secret key in base64"""
        random_bytes: bytes = urandom(32)
        return base64.b64encode(random_bytes).decode("utf-8")

    @classmethod
    def encrypt(cls, plain_text: str, secret_key: str, salt: str) -> str:
        """Return encrypted string"""
        key: bytes = base64.decodebytes(secret_key.encode("utf-8"))
        iv: bytes = base64.b64decode(salt.encode("utf-8"))
        cipher = Cipher(algorithm=algorithms.AES(key), mode=modes.CBC(iv))
        encryptor = cipher.encryptor()
        # add PKCS7 padding
        plain_bytes = plain_text.encode('utf-8')
        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(plain_bytes) + padder.finalize()
        
        encrypted_bytes = encryptor.update(padded_data) + encryptor.finalize()
        return base64.b64encode(encrypted_bytes).decode('utf-8')

    @classmethod
    def decrypt(cls, ciphertext: str, secret_key: str, salt: str) -> str:
        """Return decrypted string"""
        key: bytes = base64.decodebytes(secret_key.encode("utf-8"))
        iv: bytes = base64.decodebytes(salt.encode("utf-8"))

        encrypted_data: bytes = base64.b64decode(ciphertext)

        # AES Decryption
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
        decryptor = cipher.decryptor()
        decrypted_data = decryptor.update(encrypted_data) + decryptor.finalize()

        # Remove PKCS7 padding
        padding_length = decrypted_data[-1]
        return decrypted_data[:-padding_length].decode('utf-8')

    @classmethod
    def test(cls):
        """
        >>> Encryptor.test() # doctest: +ELLIPSIS
        Original...
        """
        # Generate a secret key (32 bytes for AES-256) and an IV (salt, 16 bytes)
        secret_key = Encryptor.generate_secret()
        salt = Encryptor.generate_salt()  # IV must be 16 bytes for AES

        original_text = "Hello, AES encryption!"
        encrypted_text = Encryptor.encrypt(original_text, secret_key, salt)
        decrypted_text = Encryptor.decrypt(encrypted_text, secret_key, salt)

        print("Original text:", original_text)
        print("Encrypted text:", encrypted_text)
        print("Decrypted text:", decrypted_text)

        # Verify that decryption returns the original text
        assert original_text == decrypted_text, "Decrypted text does not match the original"
        print("Test passed: Decrypted text matches the original!")
